@@ -32,8 +32,10 @@
 typedef struct semanage_conf {
 	enum semanage_connect_type store_type;
 	char *store_path;	/* used for both socket path and policy dir */
+	char *compiler_directory_path;
 	int server_port;
 	int policyvers;		/* version for server generated policies */
+	int target_platform;
 	int expand_check;
 	int save_previous;
 	int save_linked;
@@ -43,11 +45,15 @@ typedef struct semanage_conf {
 	mode_t file_mode;
 	int bzip_blocksize;
 	int bzip_small;
+	int remove_hll;
+	int ignore_module_cache;
+	int optimize_policy;
 	char *ignoredirs;	/* ";" separated of list for genhomedircon to ignore */
 	struct external_prog *load_policy;
 	struct external_prog *setfiles;
 	struct external_prog *sefcontext_compile;
 	struct external_prog *mod_prog, *linked_prog, *kernel_prog;
+	char *store_root_path;
 } semanage_conf_t;
 
 /* A linked list of verification programs.  Each one is called in
